import socketserver

from django.core.management.base import BaseCommand, CommandError
from django.conf import settings

from django_leek.server import TaskSocketServer


def _endpoint(endpoint):
    host, port = endpoint.split(':')
    return host, int(port)


class Command(BaseCommand):
    help = 'Starts leek worker server'

    #def add_arguments(self, parser):
    #    parser.add_argument('poll_id', nargs='+', type=int)

    def handle(self, *args, **options):
        try:
            cfg = getattr(settings, 'LEEK', {})
            host, port = _endpoint(cfg.get('bind', 'localhost:8002'))

            print('Listening on {port}'.format(port=port))
            socketserver.TCPServer.allow_reuse_address = True
            server = socketserver.TCPServer((host, port), TaskSocketServer)
            server.serve_forever()
        except KeyboardInterrupt:
            pass
        print('exiting')



import socket
from functools import wraps
from . import helpers
from .settings import HOST, PORT


class Leek(object):
    def task(self, f, pool=None):
        pool_name = pool or f.__name__

        @wraps(f)
        def _offload(*args, **kwargs):
            return push_task_to_queue(f, pool_name=pool_name, *args, **kwargs)
        f.offload = _offload
        return f


class Task(object):
    def __init__(self, a_callable, *args, **kwargs):
        assert callable(a_callable)
        self.task_callable = a_callable
        self.args = args
        self.kwargs = kwargs
    
    def __call__(self):
        self.task_callable(*self.args, **self.kwargs)


def push_task_to_queue(a_callable, *args, **kwargs):
    """Original API"""
    pool_name = kwargs.pop('pool_name', None)
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)    
    new_task = Task(a_callable, *args, **kwargs)
    queued_task = helpers.save_task_to_db(new_task, pool_name)
    sock.connect((HOST, PORT))
    sock.send("{}".format(queued_task.id).encode())
    received = sock.recv(1024)
    sock.close()

    return received



import socket
from .settings import HOST, PORT
    

def stop_server():
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.connect((HOST, PORT))
    sock.send("stop")
    received = sock.recv(1024)
    sock.close()
    print("Sent: %s" % "stop")
    print("Received: %s" % received)


if __name__ == "__main__":
    stop_server()



import pickle
import base64

from . import models


def unpack(pickled_task):
    new_task = pickle.loads(base64.b64decode(pickled_task))
    return new_task


def serialize(task):
    return base64.b64encode(pickle.dumps(task))


def load_task(task_id):
    return models.QueuedTasks.objects.get(pk=task_id)


def save_task_to_db(new_task, pool_name):
    pickled_task = serialize(new_task)
    t = models.QueuedTasks(pickled_task=pickled_task, pool=pool_name)
    t.save()
    return t



import pickle
import base64
import os
# environ settings variable, should be the same as in manage.py
os.environ.setdefault("DJANGO_SETTINGS_MODULE", "mysite.settings")
os.environ["DJANGO_SETTINGS_MODULE"] = "mysite.settings"
import django
django.setup()

 
# edit to the correct mysite.tasks_queue path
from .models import FailedTasks, QueuedTasks


Lfailed_tasks_id = FailedTasks.objects.values_list("task_id", flat=True)
tasks = QueuedTasks.objects.filter(pk__in=Lfailed_tasks_id)
for r in tasks:
    task = pickle.loads(base64.b64decode(r.pickled_task))
    task.run()



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='FailedTasks',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('task_id', models.IntegerField()),
                ('exception', models.CharField(max_length=2048)),
                ('saved_on', models.DateTimeField(auto_now_add=True)),
            ],
        ),
        migrations.CreateModel(
            name='QueuedTasks',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('pickled_task', models.CharField(max_length=5000)),
                ('queued_on', models.DateTimeField(auto_now_add=True)),
            ],
        ),
        migrations.CreateModel(
            name='SuccessTasks',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('task_id', models.IntegerField()),
                ('saved_on', models.DateTimeField(auto_now_add=True)),
            ],
        ),
    ]



from unittest.mock import patch, MagicMock
import socketserver

from django.test import TestCase
from django.core.management import call_command

from django_leek.server import TaskSocketServer
from django_leek import helpers, api


@patch.object(socketserver.TCPServer, 'serve_forever')
class LeekCommandTestCase(TestCase):
    def test_leek(self, serve_forever):
        call_command('leek')
        serve_forever.assert_called_with()

    def test_keyboard_interrupt(self, serve_forever):
        serve_forever.side_effect = KeyboardInterrupt
        call_command('leek')


def f():
    pass


class TestServer(TestCase):
    def setUp(self):
        self.request = MagicMock()

    def _request(self, data):
        if isinstance(data, Exception):
            self.request.recv.side_effect = data
        else:
            self.request.recv.return_value = data

    def _response(self):
        return b''.join(call[0][0] for call in self.request.send.call_args_list)

    def act(self):
        TaskSocketServer(self.request, 'client adress', 'server')

    def test_recv_error(self):
        self._request(OSError('Nuclear Winter'))        
        self.act()

    def test_task(self):
        task = helpers.save_task_to_db(api.Task(f), 'pool_name')
        self._request(str(task.id).encode())
        self.act()
        self.assertEqual(self._response(), b"(True, 'sent')")



# -*- coding: utf-8 -*-
# Generated by Django 1.11 on 2018-09-10 10:28
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_leek', '0002_queuedtasks_pool'),
    ]

    operations = [
        migrations.AlterField(
            model_name='queuedtasks',
            name='pickled_task',
            field=models.BinaryField(max_length=5000),
        ),
    ]



import logging
import socketserver
import multiprocessing

from .helpers import load_task
from . import worker
from . import helpers
import django

log = logging.getLogger(__name__)


Dcommands = {}


class Pool(object):
    def __init__(self):
        self.queue = multiprocessing.Queue()
        self.worker = multiprocessing.Process(target=worker.target, args=(self.queue,))


class TaskSocketServer(socketserver.BaseRequestHandler):
    DEFAULT_POOL = 'default'
    # pools holds a mapping from pool names to process objects
    pools = {}

    def handle(self):
        try:
            data = self.request.recv(5000).strip()
            if data in Dcommands.keys():
                log.info('Got command: "{}"'.format(data))
                try:
                    worker_response = Dcommands[data]()
                    response = (True, worker_response.encode(),)
                    self.request.send(str(response).encode())
                except Exception as e:
                    log.exception("command failed")
                    response =  (False, "TaskServer Command: {}".format(e).encode(),)
                    self.request.send(str(response).encode())
            else:
                # assume a serialized task
                log.info('Got a task')
                try:
                    task_id = int(data.decode())
                    django.db.connections.close_all()
                    queued_task = load_task(task_id=task_id)
                    
                    # Ensure pool got a worker processing it
                    pool_name = queued_task.pool or self.DEFAULT_POOL
                    pool = self.pools.get(pool_name)
                    if pool is None or not pool.worker.is_alive():
                        # Spawn new pool
                        log.info('Spawning new pool: {}'.format(pool_name))
                        self.pools[pool_name] = Pool()
                        self.pools[pool_name].worker.start()

                    task = helpers.unpack(queued_task.pickled_task)
                    self.pools[pool_name].queue.put(task)

                    response = (True, "sent")
                    self.request.send(str(response).encode())
                except Exception as e:
                    log.exception("failed to queue task")
                    response = (False, "TaskServer Put: {}".format(e).encode(),)
                    self.request.send(str(response).encode())

        except OSError as e:
            # in case of network error, just log
            log.exception("network error")

    def finish(self):
        for pool in self.pools.values():
            pool.queue.put(None)



from django.db import models


class QueuedTasks(models.Model):
    pickled_task = models.BinaryField(max_length=5000)  #max row 65535
    pool = models.CharField(max_length=256, null=True)
    queued_on = models.DateTimeField(auto_now_add=True)


class SuccessTasks(models.Model):
    task_id = models.IntegerField()
    saved_on = models.DateTimeField(auto_now_add=True)


class FailedTasks(models.Model):
    task_id = models.IntegerField()
    exception = models.CharField(max_length=2048)
    saved_on = models.DateTimeField(auto_now_add=True) 



import sys
from django.conf import settings


if 'test' in sys.argv:
    SECRET_KEY="just to make tests run"
    DATABASES = {'default': {'ENGINE': 'django.db.backends.sqlite3', 'NAME': ':memory:'}}
    INSTALLED_APPS = ['django_leek']


cfg = getattr(settings, "LEEK", {})
        
HOST = cfg.get('host', "localhost")
PORT = int(cfg.get('port', "8002"))



import os
from setuptools import find_packages, setup

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))


with open('README.md') as f:
    README = f.read()


setup(
    name='django-leek',
    version='0.9.2',
    packages=find_packages(exclude=['test_app']),
    install_requires = ['django>=1.11'],
    include_package_data=True,
    license='MIT License',
    description='A simple Django app to offload tasks from main web server',
    long_description=README,
    long_description_content_type='text/markdown',
    url='https://github.com/Volumental/django-leek',
    author='Volumental',
    author_email='maintainer@volumental.com',
    classifiers=[
        'Environment :: Web Environment',
        'Framework :: Django',
        'Framework :: Django :: 1.11',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
    ],
)


#  Based on the Worker class from Python in a nutshell, by Alex Martelli
import logging

import django

log = logging.getLogger(__name__)


def target(queue):
    django.setup()
    log.info('Worker Starts')
    done = False
    while not done:
        try:
            task = queue.get()
            if task is None:
                done = True
                break
                
            log.info('running task...')
            task()
            # workaround to solve problems with django + psycopg2
            # solution found here: https://stackoverflow.com/a/36580629/10385696
            django.db.connection.close()
            log.info('...successfully')
        except Exception as e:
            log.exception("...task failed")

    log.info('Worker stopped')



# -*- coding: utf-8 -*-
# Generated by Django 1.11 on 2018-09-05 03:03
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_leek', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='queuedtasks',
            name='pool',
            field=models.CharField(max_length=256, null=True),
        ),
    ]
